@@ -46,7 +46,8 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         if self.has_imported_c_functions():
             self.module_temp_cname = env.allocate_temp_pyobject()
             env.release_temp(self.module_temp_cname)
-        self.generate_dep_file(env, result)
+        if options.timestamps or options.recursive:
+            self.generate_dep_file(env, result)
         self.generate_c_code(env, result)
         self.generate_h_code(env, options, result)
         self.generate_api_code(env, result)
@@ -138,7 +139,10 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         api_funcs = []
         public_extension_types = []
         has_api_extension_types = 0
-        for entry in denv.cfunc_entries + env.cfunc_entries:
+        for entry in denv.cfunc_entries:
+            if entry.api:
+                api_funcs.append(entry)
+        for entry in env.cfunc_entries:
             if entry.api:
                 api_funcs.append(entry)
         for entry in denv.c_class_entries + env.c_class_entries:
@@ -312,7 +316,7 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.putln("  #define PyInt_FromSsize_t(z) PyInt_FromLong(z)")
         code.putln("  #define PyInt_AsSsize_t(o)	PyInt_AsLong(o)")
         code.putln("#endif")
-        code.putln("#ifndef WIN32")
+        code.putln("#if !defined(WIN32) && !defined(MS_WINDOWS)")
         code.putln("  #ifndef __stdcall")
         code.putln("    #define __stdcall")
         code.putln("  #endif")
@@ -351,8 +355,8 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.putln("#endif")
 
     def generate_includes(self, env, cimported_modules, code):
-        includes = env.include_files[:]
-        for module in cimported_modules:
+        includes = []
+        for module in cimported_modules + [env]:
             for filename in module.include_files:
                 if filename not in includes:
                     includes.append(filename)
@@ -390,13 +394,6 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
     
     def generate_type_definitions(self, env, code): #, implementation):
         #print "generate_type_definitions:", env ###
-#		if definition:
-#			type_entries = env.type_entries
-#		else:
-#			type_entries = []
-#			for entry in env.type_entries:
-#				if entry.defined_in_pxd:
-#					type_entries.append(entry)
         type_entries = env.type_entries
         self.generate_type_header_code(type_entries, code)
         for entry in env.c_class_entries:
@@ -405,21 +402,6 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 self.generate_exttype_vtable_struct(entry, code)
                 self.generate_exttype_vtabptr_declaration(entry, code)
     
-    def generate_gcc33_hack(self, env, code):
-        # Workaround for spurious warning generation in gcc 3.3
-        if 0:
-            code.putln("")
-            for entry in env.c_class_entries:
-                type = entry.type
-                if not type.typedef_flag:
-                    name = type.objstruct_cname
-                    if name.startswith("__pyx_"):
-                        tail = name[6:]
-                    else:
-                        tail = name
-                    code.putln("typedef struct %s __pyx_gcc33_%s;" % (
-                        name, tail))
-    
     def generate_typedef(self, entry, code):
         base_type = entry.type.typedef_base_type
         code.putln("")
@@ -492,9 +474,6 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 code.putln("%s DL_EXPORT(PyTypeObject) %s;" % (
                     Naming.extern_c_macro,
                     name))
-            # ??? Do we really need the rest of this? ???
-            #else:
-            #	code.putln("staticforward PyTypeObject %s;" % name)
     
     def generate_exttype_vtable_struct(self, entry, code):
         # Generate struct declaration for an extension type's vtable.
@@ -1219,7 +1198,7 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 code.putln('{"%s", %s, %s, %s, 0},' % (
                     entry.name,
                     type_code,
-                    "offsetof(%s, %s)" % (objstruct, entry.name),
+                    "offsetof(%s, %s)" % (objstruct, entry.cname),
                     flags))
             code.putln(
                     "{0, 0, 0, 0, 0}")
@@ -1248,42 +1227,6 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
             code.putln(
                 "};")
     
-#	def generate_interned_name_table(self, env, code):
-#		items = env.intern_map.items()
-#		if items:
-#			items.sort()
-#			code.putln("")
-#			code.putln(
-#				"static __Pyx_InternTabEntry %s[] = {" %
-#					Naming.intern_tab_cname)
-#			for (name, cname) in items:
-#				code.putln(
-#					'{&%s, "%s"},' % (
-#						cname,
-#						name))
-#			code.putln(
-#				"{0, 0}")
-#			code.putln(
-#				"};")
-    
-#	def generate_py_string_table(self, env, code):
-#		entries = env.all_pystring_entries
-#		if entries:
-#			code.putln("")
-#			code.putln(
-#				"static __Pyx_StringTabEntry %s[] = {" %
-#					Naming.stringtab_cname)
-#			for entry in entries:
-#				code.putln(
-#					"{&%s, %s, sizeof(%s)}," % (
-#						entry.pystring_cname,
-#						entry.cname,
-#						entry.cname))
-#			code.putln(
-#				"{0, 0, 0}")
-#			code.putln(
-#				"};")
-    
     def generate_interned_name_table(self, interned_strings, code):
         code.putln("")
         code.putln(
@@ -1293,30 +1236,6 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.putln("0")
         code.putln(
             "};")
-    
-#	def generate_py_string_table(self, py_strings, code):
-#		interned = []
-#		uninterned = []
-#		for s in py_strings:
-#			if s.is_interned:
-#				interned.append(s)
-#			else:
-#				uninterned.append(s)
-#		code.putln("")
-#		code.putln(
-#			"static __Pyx_StringTabEntry %s[] = {" %
-#				Naming.stringtab_cname)
-#		for s in interned + uninterned:
-#			cname = s.cname
-#			code.putln(
-#				"{&%s, %s, sizeof(%s)}," % (
-#					s.py_cname,
-#					cname,
-#					cname))
-#		code.putln(
-#			"{0, 0, 0}")
-#		code.putln(
-#			"};")
 
     def generate_filename_init_prototype(self, code):
         code.putln("");
@@ -1329,6 +1248,12 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         code.putln("%s; /*proto*/" % header)
         code.putln("%s {" % header)
         code.put_var_declarations(env.temp_entries)
+        
+        if env.gil_used:
+            # Workaround for GIL/threading bug in 2.3
+            code.putln("#if PY_VERSION_HEX < 0x02040000 && defined(WITH_THREAD)")
+            code.putln("  PyEval_InitThreads();")
+            code.putln("#endif")
 
         #code.putln("/*--- Libary function declarations ---*/")
         env.generate_library_function_declarations(code)
@@ -1406,13 +1331,6 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
                 Naming.builtins_cname,
                 code.error_goto(self.pos)))
     
-#	def generate_intern_code(self, env, code):
-#		code.use_utility_code(Nodes.init_intern_tab_utility_code);
-#		code.putln(
-#			"if (__Pyx_InternStrings(%s) < 0) %s;" % (
-#				Naming.intern_tab_cname,
-#				code.error_goto(self.pos)))
-    
     def generate_string_init_code(self, env, code):
         code.use_utility_code(Nodes.init_string_tab_utility_code)
         code.putln(
@@ -1431,18 +1349,14 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
     def generate_pxd_function_export_code(self, env, code):
         denv = env.definition_scope
         for entry in denv.cfunc_entries:
-            self.generate_c_function_export_code(env, entry, code)
+            if entry.visibility <> 'extern':
+                self.generate_c_function_export_code(env, entry, code)
     
     def generate_api_function_export_code(self, env, code):
         for entry in env.cfunc_entries:
             if entry.api:
                 self.generate_c_function_export_code(env, entry, code)
 
-#	def generate_c_function_export_code(self, env, code):
-#		# Generate code to create PyCFunction wrappers for exported C functions.
-#		for entry in env.cfunc_entries:
-#			if entry.api or entry.defined_in_pxd:
-    
     def generate_c_function_export_code(self, env, entry, code):
         code.use_utility_code(function_export_utility_code)
         signature = entry.type.signature_string()
@@ -1516,22 +1430,21 @@ class ModuleNode(Nodes.Node, Nodes.BlockNode):
         # extension type defined in another module, and extract its C method
         # table pointer if any.
         #print "generate_type_import_code:", type ###
-        if type in env.types_imported:
-            return
-        if type.typedef_flag:
-            objstruct = type.objstruct_cname
-        else:
-            objstruct = "struct %s" % type.objstruct_cname
-        self.generate_type_import_call(type, code, code.error_goto(pos))
-        self.use_type_import_utility_code(code)
-        if type.vtabptr_cname:
-            code.putln(
-                "if (__Pyx_GetVtable(%s->tp_dict, &%s) < 0) %s" % (
-                    type.typeptr_cname,
-                    type.vtabptr_cname,
-                    code.error_goto(pos)))
-            code.use_utility_code(Nodes.get_vtable_utility_code)
-        env.types_imported[type] = 1
+        if not type.is_builtin and type not in env.types_imported:
+            if type.typedef_flag:
+                objstruct = type.objstruct_cname
+            else:
+                objstruct = "struct %s" % type.objstruct_cname
+            self.generate_type_import_call(type, code, code.error_goto(pos))
+            self.use_type_import_utility_code(code)
+            if type.vtabptr_cname:
+                code.putln(
+                    "if (__Pyx_GetVtable(%s->tp_dict, &%s) < 0) %s" % (
+                        type.typeptr_cname,
+                        type.vtabptr_cname,
+                        code.error_goto(pos)))
+                code.use_utility_code(Nodes.get_vtable_utility_code)
+            env.types_imported[type] = 1
     
     def generate_type_import_call(self, type, code, error_code):
         if type.typedef_flag:
@@ -1651,7 +1564,7 @@ bad:
 type_import_utility_code = [
 """
 static PyTypeObject *__Pyx_ImportType(char *module_name, char *class_name, long size);  /*proto*/
-""","""
+""",r"""
 #ifndef __PYX_HAVE_RT_ImportType
 #define __PYX_HAVE_RT_ImportType
 static PyTypeObject *__Pyx_ImportType(char *module_name, char *class_name, 
@@ -1710,6 +1623,7 @@ static int __Pyx_ExportFunction(char *n, void *f, char *s) {
         goto bad;
     if (PyDict_SetItemString(d, n, p) < 0)
         goto bad;
+    Py_DECREF(p);
     Py_DECREF(d);
     return 0;
 bad:
