@@ -21,7 +21,7 @@ class BaseType:
 
 class PyrexType(BaseType):
     #
-    #  Base class for all Pyrex types.
+    #  Base class for all non-pseudo Pyrex types.
     #
     #  is_pyobject           boolean     Is a Python object type
     #  is_extension_type     boolean     Is a Python extension type
@@ -38,6 +38,8 @@ class PyrexType(BaseType):
     #  is_typedef            boolean     Is a typedef type
     #  is_string             boolean     Is a C char * type
     #  is_returncode         boolean     Is used only to signal exceptions
+    #  is_sequence           boolean     Is a sequence type
+    #  is_builtin            boolean     Is a built-in Python type
     #  is_error              boolean     Is the dummy error type
     #  has_attributes        boolean     Has C dot-selectable attributes
     #  default_value         string      Initial value
@@ -84,6 +86,8 @@ class PyrexType(BaseType):
     is_typedef = 0
     is_string = 0
     is_returncode = 0
+    is_sequence = 0
+    is_builtin = 0
     is_error = 0
     has_attributes = 0
     default_value = ""
@@ -129,7 +133,26 @@ class PyrexType(BaseType):
         return 1
 
 
-class CTypedefType(BaseType):
+class TypeWrapper(BaseType):
+    #  Base class for pseudo-types that delegate most
+    #  attribute lookups to another type.
+    #
+    #  delegate_type   PyrexType
+    
+    def __init__(self, base_type):
+        self.delegate_type = base_type
+
+    def __getattr__(self, name):
+        return getattr(self.delegate_type, name)
+    
+    def define(self, base_type):
+        self.delegate_type = base_type
+
+    def resolve(self):
+        return self.delegate_type.resolve()
+    
+
+class CTypedefType(TypeWrapper):
     #
     #  Pseudo-type defined with a ctypedef statement in a
     #  'cdef extern from' block. Delegates most attribute
@@ -143,12 +166,10 @@ class CTypedefType(BaseType):
     is_typedef = 1
     
     def __init__(self, cname, base_type):
+        TypeWrapper.__init__(self, base_type)
         self.typedef_cname = cname
         self.typedef_base_type = base_type
     
-    def resolve(self):
-        return self.typedef_base_type.resolve()
-    
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
         name = self.declaration_name(for_display, pyrex)
@@ -169,9 +190,6 @@ class CTypedefType(BaseType):
     def __str__(self):
         return self.declaration_name(for_display = 1)
     
-    def __getattr__(self, name):
-        return getattr(self.typedef_base_type, name)
-
 
 class PyObjectType(PyrexType):
     #
@@ -234,6 +252,8 @@ class PyExtensionType(PyObjectType):
         self.vtabstruct_cname = None
         self.vtabptr_cname = None
         self.vtable_cname = None
+        if base_type and base_type.is_sequence:
+            self.is_sequence = 1
     
     def set_scope(self, scope):
         self.scope = scope
@@ -271,6 +291,10 @@ class PyExtensionType(PyObjectType):
     def attributes_known(self):
         return self.scope is not None
     
+    def is_defined(self):
+        scope = self.scope
+        return scope and (scope.defined or scope.implemented)
+    
     def __str__(self):
         return self.name
     
@@ -484,14 +508,18 @@ class CPtrType(CType):
     def assignable_from_resolved_type(self, other_type):
         if other_type is error_type:
             return 1
-        elif other_type.is_null_ptr:
+        if other_type.is_null_ptr:
             return 1
-        elif self.base_type.is_cfunction and other_type.is_cfunction:
-            return self.base_type.same_as(other_type)
-        elif other_type.is_array or other_type.is_ptr:
+        if self.base_type.is_cfunction:
+            if other_type.is_ptr:
+                other_type = other_type.base_type.resolve()
+            if other_type.is_cfunction:
+                return self.base_type.pointer_assignable_from_resolved_type(other_type)
+            else:
+                return 0
+        if other_type.is_array or other_type.is_ptr:
             return self.base_type.is_void or self.base_type.same_as(other_type.base_type)
-        else:
-            return 0
+        return 0
 
 
 class CNullPtrType(CPtrType):
@@ -543,7 +571,9 @@ class CFuncType(CType):
         return self.same_c_signature_as_resolved_type(
             other_type.resolve(), as_cmethod)
 
-    def same_c_signature_as_resolved_type(self, other_type, as_cmethod):
+    def same_c_signature_as_resolved_type(self, other_type, as_cmethod = 0):
+        #print "CFuncType.same_c_signature_as_resolved_type:", \
+        #	self, other_type, "as_cmethod =", as_cmethod ###
         if other_type is error_type:
             return 1
         if not other_type.is_cfunction:
@@ -581,7 +611,13 @@ class CFuncType(CType):
     
     def same_as_resolved_type(self, other_type, as_cmethod = 0):
         return self.same_c_signature_as_resolved_type(other_type, as_cmethod) \
-            and self.same_exception_signature_as_resolved_type(other_type)
+            and self.same_exception_signature_as_resolved_type(other_type) \
+            and self.nogil == other_type.nogil
+    
+    def pointer_assignable_from_resolved_type(self, other_type):
+        return self.same_c_signature_as_resolved_type(other_type) \
+            and self.same_exception_signature_as_resolved_type(other_type) \
+            and not (self.nogil and not other_type.nogil)
     
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
@@ -594,21 +630,27 @@ class CFuncType(CType):
         arg_decl_code = string.join(arg_decl_list, ",")
         if not arg_decl_code and not pyrex:
             arg_decl_code = "void"
-        exc_clause = ""
+        trailer = ""
         if (pyrex or for_display) and not self.return_type.is_pyobject:
             if self.exception_value and self.exception_check:
-                exc_clause = " except? %s" % self.exception_value
+                trailer = " except? %s" % self.exception_value
             elif self.exception_value:
-                exc_clause = " except %s" % self.exception_value
+                trailer = " except %s" % self.exception_value
             elif self.exception_check:
-                exc_clause = " except *"
+                trailer = " except *"
+            if self.nogil:
+                trailer += " nogil"
         cc = self.calling_convention_prefix()
         if (not entity_code and cc) or entity_code.startswith("*"):
             entity_code = "(%s%s)" % (cc, entity_code)
             cc = ""
         return self.return_type.declaration_code(
-            "%s%s(%s)%s" % (cc, entity_code, arg_decl_code, exc_clause),
+            "%s%s(%s)%s" % (cc, entity_code, arg_decl_code, trailer),
             for_display, dll_linkage, pyrex)
+    
+    def function_header_code(self, func_name, arg_code):
+        return "%s%s(%s)" % (self.calling_convention_prefix(),
+            func_name, arg_code)
 
     def signature_string(self):
         s = self.declaration_code("")
@@ -751,6 +793,7 @@ class ErrorType(PyrexType):
     exception_check	= 0
     to_py_function = "dummy"
     from_py_function = "dummy"
+    parsetuple_format = "E"
     
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
@@ -761,6 +804,7 @@ class ErrorType(PyrexType):
 
 
 py_object_type = PyObjectType()
+py_type_type = TypeWrapper(None) # Bootstrapping placeholder, filled later
 
 c_void_type =         CVoidType()
 c_void_ptr_type =     CPtrType(c_void_type)
@@ -912,8 +956,8 @@ def typecast(to_type, from_type, expr_code):
     #  assigned to to_type, assuming its existing C type
     #  is from_type.
     if to_type is from_type or \
+        same_type(to_type, from_type) or \
         (not to_type.is_pyobject and assignable_from(to_type, from_type)):
             return expr_code
     else:
-        #print "typecast: to", to_type, "from", from_type ###
         return to_type.cast_code(expr_code)
